package saferepo.model;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;

public class MyRepository extends AbstractRepository {

	public MyRepository() {
		super();
	}
	
	@Override
	public VersionedDocument add(Document document, LocalDateTime timestamp) {
		String documentID = document.getID();
			// System.out.println("DEBUG: VersionedDocument: " + document);
		List<VersionedDocument> versions = repomap.get(documentID);
		if (versions==null) versions = new ArrayList<>();
		VersionedDocument vdoc = new VersionedDocument(document, timestamp, versions.size());
		versions.add(vdoc);
		repomap.put(documentID, versions);
		return vdoc;
	}
	
	@Override
	public boolean delete(String documentID) {
		List<VersionedDocument> versions = repomap.get(documentID);
		if (versions==null) return false;
		else {
			versions.add(new VersionedDocument(new Document(documentID, null), LocalDateTime.now(), versions.size()));
			repomap.put(documentID, versions);
			return true;
		}
	}

	@Override
	public Optional<VersionedDocument> getCurrentVersion(String documentID) {
		List<VersionedDocument> versions = repomap.get(documentID);
		return (versions==null || versions.isEmpty()) ? Optional.empty() : Optional.ofNullable(versions.get(versions.size()-1));
	}
	
	@Override
	public Optional<VersionedDocument> getVersion(String documentID, int n){
		List<VersionedDocument> versions = repomap.get(documentID);
		if (n<0) throw new IllegalArgumentException();
		if (n>= versions.size()) return Optional.empty();
		return (versions==null || versions.isEmpty()) ? Optional.empty() : Optional.of(versions.get(n));
	}

	@Override
	public Optional<VersionedDocument> getVersionAt(String documentID, LocalDateTime timestamp) {
		List<VersionedDocument> versions = repomap.get(documentID);
		return (versions==null || versions.isEmpty()) ? Optional.empty() : 
				versions.stream().filter(
						doc -> doc.getTimeStamp().isBefore(timestamp) || doc.getTimeStamp().equals(timestamp) 
						).max(Comparator.comparing(VersionedDocument::getTimeStamp)) ;
	}	
	
	/*
	public static void main(String[] args) throws InterruptedException {
		Document fakeDoc1 = new Document("doc1", Paths.get("/temp/doc1"));
		Document fakeDoc2 = new Document("doc2", Paths.get("/temp/doc2"));
		Document fakeDoc3 = new Document("doc3", Paths.get("/temp/doc3"));

		Repository repo = new MyRepository();
		
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:39:02"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:39:03"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:40:04"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:40:05"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:42:06"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:42:07"));
		repo.add(fakeDoc1, LocalDateTime.parse("2018-12-30T09:44:08"));

		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:39:04"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:39:05"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:40:04"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:40:08"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:41:06"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:41:07"));
		repo.add(fakeDoc2, LocalDateTime.parse("2018-12-30T09:43:08"));
		
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:39:01"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:39:05"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:40:06"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:40:08"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:40:09"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:41:10"));
		repo.add(fakeDoc3, LocalDateTime.parse("2018-12-30T09:44:09"));

		
		System.out.println("Current version of doc1: " + repo.getCurrentVersion("doc1").get());
		System.out.println("Current version of doc2: " + repo.getCurrentVersion("doc2").get());
		System.out.println("Current version of doc3: " + repo.getCurrentVersion("doc3").get());
		
		System.out.println("Previous version #2 of doc3: " +  repo.getVersion("doc3", 2) );
		System.out.println("Previous version #3 of doc3: " +  repo.getVersion("doc3", 3) );
		System.out.println("Version of doc3 valid at 2018-12-30T09:40:06: " + repo.getVersionAt("doc3", LocalDateTime.parse("2018-12-30T09:40:06")) );
		System.out.println("Version of doc3 valid at 2018-12-30T09:40:07: " + repo.getVersionAt("doc3", LocalDateTime.parse("2018-12-30T09:40:07")) );
		
		System.out.println("Previous version #4 of doc1: " +  repo.getVersion("doc1", 4) );
		System.out.println("Restoring version #4 of doc1: " +  repo.restoreVersion("doc1", 4) ); // true, ma produce la v7
		System.out.println("Current version of doc1: " + repo.getCurrentVersion("doc1").get());  // v7
		System.out.print(  "Checking doc1 contents @v4 vs. v7: " );
		System.out.println(repo.getVersion("doc1", 4).get().getDocument()==repo.getVersion("doc1", 7).get().getDocument());	// true, perché v7 e v4 incapsulano lo stesso doc
		
		System.out.println("Deleting doc2: " + repo.delete("doc2"));
		System.out.println("Version of doc2 valid now: " + repo.getVersionAt("doc2", LocalDateTime.now()) );     // inner doc ==null
		System.out.println("Restoring doc2 valid 1ms ago: " +  repo.restoreVersionAt("doc2", LocalDateTime.now().minusNanos(1000)) ); // giustamente non lo fa
		System.out.println("Previous version #6 of doc2: " +  repo.getVersion("doc2", 6) );
		System.out.println("Restoring doc2 valid 1 min ago: "  +  repo.restoreVersionAt("doc2", LocalDateTime.now().minusMinutes(1)) );  // ora invece lo fa, e trova la v7
		
		System.out.println("Current version of doc1: " + repo.getCurrentVersion("doc1").get());
		System.out.println("Current version of doc2: " + repo.getCurrentVersion("doc2").get());
		System.out.println("Current version of doc3: " + repo.getCurrentVersion("doc3").get());
		
	}
	*/

}
